# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging
from collections import UserList
from typing import TYPE_CHECKING, Annotated, Any, Iterable, Optional, Union

from shapely.geometry import GeometryCollection, shape
from typing_extensions import Doc

from eodag.api.product import EOProduct, unregistered_product_from_item
from eodag.plugins.crunch.filter_date import FilterDate
from eodag.plugins.crunch.filter_latest_intersect import FilterLatestIntersect
from eodag.plugins.crunch.filter_latest_tpl_name import FilterLatestByName
from eodag.plugins.crunch.filter_overlap import FilterOverlap
from eodag.plugins.crunch.filter_property import FilterProperty
from eodag.utils import GENERIC_STAC_PROVIDER, STAC_SEARCH_PLUGINS
from eodag.utils.exceptions import MisconfiguredError

if TYPE_CHECKING:
    from shapely.geometry.base import BaseGeometry

    from eodag.plugins.crunch.base import Crunch
    from eodag.plugins.manager import PluginManager


logger = logging.getLogger("eodag.search_result")


class SearchResult(UserList[EOProduct]):
    """An object representing a collection of :class:`~eodag.api.product._product.EOProduct` resulting from a search.

    :param products: A list of products resulting from a search
    :param number_matched: (optional) the estimated total number of matching results

    :cvar data: List of products
    :ivar number_matched: Estimated total number of matching results
    """

    errors: Annotated[
        list[tuple[str, Exception]], Doc("Tuple of provider name, exception")
    ]

    def __init__(
        self,
        products: list[EOProduct],
        number_matched: Optional[int] = None,
        errors: Optional[list[tuple[str, Exception]]] = None,
    ) -> None:
        super().__init__(products)
        self.number_matched = number_matched
        self.errors = errors if errors is not None else []

    def crunch(self, cruncher: Crunch, **search_params: Any) -> SearchResult:
        """Do some crunching with the underlying EO products.

        :param cruncher: The plugin instance to use to work on the products
        :param search_params: The criteria that have been used to produce this result
        :returns: The result of the application of the crunching method to the EO products
        """
        crunched_results = cruncher.proceed(self.data, **search_params)
        return SearchResult(crunched_results)

    def filter_date(
        self, start: Optional[str] = None, end: Optional[str] = None
    ) -> SearchResult:
        """
        Apply :class:`~eodag.plugins.crunch.filter_date.FilterDate` crunch,
        check its documentation to know more.
        """
        return self.crunch(FilterDate(dict(start=start, end=end)))

    def filter_latest_intersect(
        self, geometry: Union[dict[str, Any], BaseGeometry, Any]
    ) -> SearchResult:
        """
        Apply :class:`~eodag.plugins.crunch.filter_latest_intersect.FilterLatestIntersect` crunch,
        check its documentation to know more.
        """
        return self.crunch(FilterLatestIntersect({}), geometry=geometry)

    def filter_latest_by_name(self, name_pattern: str) -> SearchResult:
        """
        Apply :class:`~eodag.plugins.crunch.filter_latest_tpl_name.FilterLatestByName` crunch,
        check its documentation to know more.
        """
        return self.crunch(FilterLatestByName(dict(name_pattern=name_pattern)))

    def filter_overlap(
        self,
        geometry: Any,
        minimum_overlap: int = 0,
        contains: bool = False,
        intersects: bool = False,
        within: bool = False,
    ) -> SearchResult:
        """
        Apply :class:`~eodag.plugins.crunch.filter_overlap.FilterOverlap` crunch,
        check its documentation to know more.
        """
        return self.crunch(
            FilterOverlap(
                dict(
                    minimum_overlap=minimum_overlap,
                    contains=contains,
                    intersects=intersects,
                    within=within,
                )
            ),
            geometry=geometry,
        )

    def filter_property(
        self, operator: str = "eq", **search_property: Any
    ) -> SearchResult:
        """
        Apply :class:`~eodag.plugins.crunch.filter_property.FilterProperty` crunch,
        check its documentation to know more.
        """
        return self.crunch(FilterProperty(dict(operator=operator, **search_property)))

    def filter_online(self) -> SearchResult:
        """
        Use cruncher :class:`~eodag.plugins.crunch.filter_property.FilterProperty`,
        filter for online products.
        """
        return self.filter_property(storageStatus="ONLINE")

    @staticmethod
    def from_geojson(feature_collection: dict[str, Any]) -> SearchResult:
        """Builds an :class:`~eodag.api.search_result.SearchResult` object from its representation as geojson

        :param feature_collection: A collection representing a search result.
        :returns: An eodag representation of a search result
        """
        return SearchResult(
            [
                EOProduct.from_geojson(feature)
                for feature in feature_collection["features"]
            ]
        )

    def as_geojson_object(self) -> dict[str, Any]:
        """GeoJSON representation of SearchResult"""
        return {
            "type": "FeatureCollection",
            "features": [product.as_dict() for product in self],
        }

    def as_shapely_geometry_object(self) -> GeometryCollection:
        """:class:`shapely.GeometryCollection` representation of SearchResult"""
        return GeometryCollection(
            [
                shape(feature["geometry"]).buffer(0)
                for feature in self.as_geojson_object()["features"]
            ]
        )

    def as_wkt_object(self) -> str:
        """WKT representation of SearchResult"""
        return self.as_shapely_geometry_object().wkt

    @property
    def __geo_interface__(self) -> dict[str, Any]:
        """Implements the geo-interface protocol.

        See https://gist.github.com/sgillies/2217756
        """
        return self.as_geojson_object()

    def _repr_html_(self):
        total_count = f"/{self.number_matched}" if self.number_matched else ""
        return (
            f"""<table>
                <thead><tr><td style='text-align: left; color: grey;'>
                 {type(self).__name__}&ensp;({len(self)}{total_count})
                </td></tr></thead>
            """
            + "".join(
                [
                    f"""<tr><td style='text-align: left;'>
                <details><summary style='color: grey; font-family: monospace;'>
                    {i}&ensp;
                    {type(p).__name__}(id=<span style='color: black;'>{
                        p.properties["id"]
                    }</span>, provider={p.provider})
                </summary>
                {p._repr_html_()}
                </details>
                </td></tr>
                """
                    for i, p in enumerate(self)
                ]
            )
            + "</table>"
        )

    def extend(self, other: Iterable) -> None:
        """override extend method to include errors"""
        if isinstance(other, SearchResult):
            self.errors.extend(other.errors)

        return super().extend(other)

    @classmethod
    def _from_stac_item(
        cls, feature: dict[str, Any], plugins_manager: PluginManager
    ) -> SearchResult:
        """Create a SearchResult from a STAC item.

        :param feature: A STAC item as a dictionary
        :param plugins_manager: The EODAG plugin manager instance
        :returns: A SearchResult containing the EOProduct(s) created from the STAC item
        """
        # Try importing from EODAG Server
        if results := _import_stac_item_from_eodag_server(feature, plugins_manager):
            return results

        # try importing from a known STAC provider
        if results := _import_stac_item_from_known_provider(feature, plugins_manager):
            return results

        # try importing from an unknown STAC provider
        return _import_stac_item_from_unknown_provider(feature, plugins_manager)


def _import_stac_item_from_eodag_server(
    feature: dict[str, Any], plugins_manager: PluginManager
) -> Optional[SearchResult]:
    """Import a STAC item from EODAG Server.

    :param feature: A STAC item as a dictionary
    :param plugins_manager: The EODAG plugin manager instance
    :returns: A SearchResult containing the EOProduct(s) created from the STAC item
    """
    provider = None
    if backends := feature["properties"].get("federation:backends"):
        provider = backends[0]
    elif providers := feature["properties"].get("providers"):
        provider = providers[0].get("name")
    if provider is not None:
        logger.debug("Trying to import STAC item from EODAG Server")
        # assets coming from a STAC provider
        assets = {
            k: v["alternate"]["origin"]
            for k, v in feature.get("assets", {}).items()
            if k not in ("thumbnail", "downloadLink")
            and "origin" in v.get("alternate", {})
        }
        if assets:
            updated_item = {**feature, **{"assets": assets}}
        else:
            # item coming from a non-STAC provider
            updated_item = {**feature}
            download_link = (
                feature.get("assets", {})
                .get("downloadLink", {})
                .get("alternate", {})
                .get("origin", {})
                .get("href")
            )
            if download_link:
                updated_item["assets"] = {}
                updated_item["links"] = [{"rel": "self", "href": download_link}]
            else:
                updated_item = {}
        try:
            eo_product = unregistered_product_from_item(
                updated_item, GENERIC_STAC_PROVIDER, plugins_manager
            )
        except MisconfiguredError:
            eo_product = None
        if eo_product is not None:
            eo_product.provider = provider
            eo_product._register_downloader_from_manager(plugins_manager)
            return SearchResult([eo_product])
    return None


def _import_stac_item_from_known_provider(
    feature: dict[str, Any], plugins_manager: PluginManager
) -> Optional[SearchResult]:
    """Import a STAC item from an already-configured STAC provider.

    :param feature: A STAC item as a dictionary
    :param plugins_manager: The EODAG plugin manager instance
    :returns: A SearchResult containing the EOProduct(s) created from the STAC item
    """
    item_hrefs = [f for f in feature.get("links", []) if f.get("rel") == "self"]
    item_href = item_hrefs[0]["href"] if len(item_hrefs) > 0 else None
    imported_products = SearchResult([])
    for search_plugin in plugins_manager.get_search_plugins():
        # only try STAC search plugins
        if (
            search_plugin.config.type in STAC_SEARCH_PLUGINS
            and search_plugin.provider != GENERIC_STAC_PROVIDER
            and hasattr(search_plugin, "normalize_results")
        ):
            provider_base_url = search_plugin.config.api_endpoint.removesuffix(
                "/search"
            )
            # compare the item href with the provider base URL
            if item_href and item_href.startswith(provider_base_url):
                products = search_plugin.normalize_results([feature])
                if len(products) == 0 or len(products[0].assets) == 0:
                    continue
                logger.debug(
                    "Trying to import STAC item from %s", search_plugin.provider
                )
                eo_product = products[0]

                configured_pts = [
                    k
                    for k, v in search_plugin.config.products.items()
                    if v.get("productType") == feature.get("collection")
                ]
                if len(configured_pts) > 0:
                    eo_product.product_type = configured_pts[0]
                else:
                    eo_product.product_type = feature.get("collection")

                eo_product._register_downloader_from_manager(plugins_manager)
                imported_products.append(eo_product)
    if len(imported_products) > 0:
        return imported_products
    return None


def _import_stac_item_from_unknown_provider(
    feature: dict[str, Any], plugins_manager: PluginManager
) -> SearchResult:
    """Import a STAC item from an unknown STAC provider.

    :param feature: A STAC item as a dictionary
    :param plugins_manager: The EODAG plugin manager instance
    :returns: A SearchResult containing the EOProduct(s) created from the STAC item
    """
    try:
        logger.debug("Trying to import STAC item from unknown provider")
        eo_product = unregistered_product_from_item(
            feature, GENERIC_STAC_PROVIDER, plugins_manager
        )
    except MisconfiguredError:
        pass
    if eo_product is not None:
        eo_product.product_type = feature.get("collection")
        eo_product._register_downloader_from_manager(plugins_manager)
        return SearchResult([eo_product])
    else:
        return SearchResult([])


class RawSearchResult(UserList[dict[str, Any]]):
    """An object representing a collection of raw/unparsed search results obtained from a provider.

    :param results: A list of raw/unparsed search results
    """

    query_params: dict[str, Any]
    product_type_def_params: dict[str, Any]

    def __init__(self, results: list[Any]) -> None:
        super(RawSearchResult, self).__init__(results)



# -*- coding: utf-8 -*-
# Copyright 2024, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import copy
import logging
import os
import re
from datetime import datetime
from typing import TYPE_CHECKING, Any, Optional, cast
from urllib.parse import urlsplit

import boto3
import botocore
import requests
from dateutil.parser import isoparse
from dateutil.tz import tzutc
from dateutil.utils import today

from eodag import EOProduct
from eodag.api.product import AssetsDict
from eodag.config import PluginConfig
from eodag.plugins.search import PreparedSearch
from eodag.plugins.search.static_stac_search import StaticStacSearch
from eodag.utils import get_bucket_name_and_prefix, get_geometry_from_various
from eodag.utils.exceptions import RequestError, UnsupportedProductType, ValidationError

if TYPE_CHECKING:
    from mypy_boto3_s3 import S3Client
    from mypy_boto3_s3.type_defs import ListObjectsOutputTypeDef

logger = logging.getLogger("eodag.search.cop_marine")


def _get_date_from_yyyymmdd(date_str: str, item_key: str) -> Optional[datetime]:
    year = date_str[:4]
    month = date_str[4:6]
    if len(date_str) > 6:
        day = date_str[6:]
    else:
        day = "1"
    try:
        date = datetime(
            int(year),
            int(month),
            int(day),
            tzinfo=tzutc(),
        )
    except ValueError:
        logger.error(f"{item_key}: {date_str} is not a valid date")
        return None
    else:
        return date


def _get_dates_from_dataset_data(
    dataset_item: dict[str, Any]
) -> Optional[dict[str, str]]:
    dates = {}
    if "start_datetime" in dataset_item["properties"]:
        dates["start"] = dataset_item["properties"]["start_datetime"]
        dates["end"] = dataset_item["properties"]["end_datetime"]
    elif "datetime" in dataset_item["properties"]:
        dates["start"] = dataset_item["properties"]["datetime"]
        dates["end"] = dataset_item["properties"]["datetime"]
    else:
        return None
    return dates


def _get_s3_client(endpoint_url: str) -> S3Client:
    s3_session = boto3.Session()
    return s3_session.client(
        "s3",
        config=botocore.config.Config(
            # Configures to use subdomain/virtual calling format.
            s3={"addressing_style": "virtual"},
            signature_version=botocore.UNSIGNED,
        ),
        endpoint_url=endpoint_url,
    )


def _check_int_values_properties(properties: dict[str, Any]):
    # remove int values with a bit length of more than 64 from the properties
    invalid = []
    for prop, prop_value in properties.items():
        if isinstance(prop_value, int) and prop_value.bit_length() > 64:
            invalid.append(prop)
        if isinstance(prop_value, dict):
            _check_int_values_properties(prop_value)

    for inv_key in invalid:
        properties.pop(inv_key)


class CopMarineSearch(StaticStacSearch):
    """class that implements search for the Copernicus Marine provider

    It calls :meth:`~eodag.plugins.search.static_stac_search.StaticStacSearch.discover_product_types`
    inherited from :class:`~eodag.plugins.search.static_stac_search.StaticStacSearch`
    but for the actual search a special method which fetches the urls of the available products from an S3 storage and
    filters them has been written.

    The configuration parameters are inherited from the parent and grand-parent classes. The
    :attr:`~eodag.config.PluginConfig.DiscoverMetadata.auto_discovery` parameter in the
    :attr:`~eodag.config.PluginConfig.discover_metadata` section has to be set to ``false`` and the
    :attr:`~eodag.config.PluginConfig.DiscoverQueryables.fetch_url` in the
    :attr:`~eodag.config.PluginConfig.discover_queryables` queryables section has to be set to ``null`` to
    overwrite the default config from the stac provider configuration because those functionalities
    are not available.
    """

    def __init__(self, provider: str, config: PluginConfig):
        original_metadata_mapping = copy.deepcopy(config.metadata_mapping)
        super().__init__(provider, config)
        # reset to original metadata mapping from config (changed in super class init)
        self.config.metadata_mapping = original_metadata_mapping

    def _get_product_type_info(
        self, product_type: str
    ) -> tuple[dict[str, Any], list[dict[str, Any]]]:
        """Fetch product type and associated datasets info"""

        fetch_url = cast(str, self.config.discover_product_types["fetch_url"]).format(
            **self.config.__dict__
        )

        logger.debug("fetch data for collection %s", product_type)
        provider_product_type = self.config.products.get(product_type, {}).get(
            "productType", None
        )
        if not provider_product_type:
            provider_product_type = product_type
        collection_url = (
            fetch_url.replace("catalog.stac.json", provider_product_type)
            + "/product.stac.json"
        )
        try:
            collection_data = requests.get(collection_url).json()
        except requests.RequestException as exc:
            if exc.errno == 404:
                logger.error("product %s not found", product_type)
                raise UnsupportedProductType(product_type)
            logger.error("data for product %s could not be fetched", product_type)
            raise RequestError.from_error(
                exc, f"data for product {product_type} could not be fetched"
            ) from exc

        datasets = []
        for link in [li for li in collection_data["links"] if li["rel"] == "item"]:
            dataset_url = (
                fetch_url.replace("catalog.stac.json", provider_product_type)
                + "/"
                + link["href"]
            )
            try:
                dataset_item = requests.get(dataset_url).json()
                datasets.append(dataset_item)
            except requests.RequestException:
                logger.error("data for dataset %s could not be fetched", link["title"])

        return collection_data, datasets

    def _get_product_by_id(
        self,
        collection_objects: ListObjectsOutputTypeDef,
        product_id: str,
        s3_url: str,
        product_type: str,
        dataset_item: dict[str, Any],
        collection_dict: dict[str, Any],
    ):
        # try to find date(s) in product id
        item_dates = re.findall(r"(\d{4})(0[1-9]|1[0-2])([0-3]\d)", product_id)
        if not item_dates:
            item_dates = re.findall(r"_(\d{4})(0[1-9]|1[0-2])", product_id)
        use_dataset_dates = not bool(item_dates)
        for obj in collection_objects["Contents"]:
            if product_id in obj["Key"]:
                return self._create_product(
                    product_type,
                    obj["Key"],
                    s3_url,
                    dataset_item,
                    collection_dict,
                    use_dataset_dates,
                )
        return None

    def _create_product(
        self,
        product_type: str,
        item_key: str,
        s3_url: str,
        dataset_item: dict[str, Any],
        collection_dict: dict[str, Any],
        use_dataset_dates: bool = False,
        product_id: Optional[str] = None,
    ) -> Optional[EOProduct]:

        item_id = os.path.splitext(item_key.split("/")[-1])[0]
        if product_id and product_id != item_id:
            return None
        download_url = s3_url + "/" + item_key
        geometry = (
            get_geometry_from_various(**dataset_item)
            or self.config.metadata_mapping["defaultGeometry"]
        )
        properties = {
            "id": item_id,
            "title": item_id,
            "geometry": geometry,
            "downloadLink": download_url,
            "dataset": dataset_item["id"],
        }
        if use_dataset_dates:
            dates = _get_dates_from_dataset_data(dataset_item)
            if not dates:
                return None
            properties["startTimeFromAscendingNode"] = dates["start"]
            properties["completionTimeFromAscendingNode"] = dates["end"]
        else:
            item_dates = re.findall(r"(\d{4})(0[1-9]|1[0-2])([0-3]\d)", item_id)
            if not item_dates:
                item_dates = re.findall(r"_(\d{4})(0[1-9]|1[0-2])", item_id)
            item_dates = ["".join(row) for row in item_dates]
            item_start = _get_date_from_yyyymmdd(item_dates[0], item_key)
            if not item_start:  # identified pattern was not a valid datetime
                return None
            if len(item_dates) > 2:  # start, end and created_at timestamps
                item_end = _get_date_from_yyyymmdd(item_dates[1], item_key)
            else:  # only date and created_at timestamps
                item_end = item_start
            properties["startTimeFromAscendingNode"] = item_start.strftime(
                "%Y-%m-%dT%H:%M:%SZ"
            )
            properties["completionTimeFromAscendingNode"] = (
                item_end or item_start
            ).strftime("%Y-%m-%dT%H:%M:%SZ")

        for key, value in collection_dict["properties"].items():
            if key not in ["id", "title", "start_datetime", "end_datetime", "datetime"]:
                properties[key] = value
        for key, value in dataset_item["properties"].items():
            if key not in ["id", "title", "start_datetime", "end_datetime", "datetime"]:
                properties[key] = value

        code_mapping = self.config.products.get(product_type, {}).get(
            "code_mapping", None
        )
        if code_mapping:
            id_parts = item_id.split("_")
            if len(id_parts) > code_mapping["index"]:
                code = id_parts[code_mapping["index"]]
                if "pattern" not in code_mapping:
                    properties[code_mapping["param"]] = code
                elif re.findall(code_mapping["pattern"], code):
                    properties[code_mapping["param"]] = re.findall(
                        code_mapping["pattern"], code
                    )[0]

        _check_int_values_properties(properties)

        properties["thumbnail"] = collection_dict["assets"]["thumbnail"]["href"]
        if "omiFigure" in collection_dict["assets"]:
            properties["quicklook"] = collection_dict["assets"]["omiFigure"]["href"]
        assets = {
            "native": {
                "title": "native",
                "href": download_url,
                "type": "application/x-netcdf",
            }
        }
        additional_assets = self.get_assets_from_mapping(dataset_item)
        assets.update(additional_assets)
        product = EOProduct(self.provider, properties, productType=product_type)
        # use product_type_config as default properties
        product_type_config = getattr(self.config, "product_type_config", {})
        product.properties = dict(product_type_config, **product.properties)
        product.assets = AssetsDict(product, assets)
        return product

    def query(
        self,
        prep: PreparedSearch = PreparedSearch(),
        **kwargs: Any,
    ) -> tuple[list[EOProduct], Optional[int]]:
        """
        Implementation of search for the Copernicus Marine provider
        :param prep: object containing search parameterds
        :param kwargs: additional search arguments
        :returns: list of products and total number of products
        """
        page = prep.page
        items_per_page = prep.items_per_page

        # only return 1 page if pagination is disabled
        if page is None or items_per_page is None or page > 1 and items_per_page <= 0:
            return ([], 0) if prep.count else ([], None)

        product_type = kwargs.get("productType", prep.product_type)
        if not product_type:
            raise ValidationError(
                "parameter product type is required for search with cop_marine provider"
            )
        collection_dict, datasets_items_list = self._get_product_type_info(product_type)
        geometry = kwargs.pop("geometry", None)
        products: list[EOProduct] = []
        start_index = items_per_page * (page - 1) + 1
        num_total = 0
        for i, dataset_item in enumerate(datasets_items_list):
            # Filter by geometry
            if "id" not in kwargs and geometry:
                dataset_geom = get_geometry_from_various(**dataset_item)
                if dataset_geom and not dataset_geom.intersects(geometry):
                    continue
            try:
                logger.debug("searching data for dataset %s", dataset_item["id"])

                # date bounds
                if "startTimeFromAscendingNode" in kwargs:
                    start_date = isoparse(kwargs["startTimeFromAscendingNode"])
                elif "start_datetime" in dataset_item["properties"]:
                    start_date = isoparse(dataset_item["properties"]["start_datetime"])
                else:
                    start_date = isoparse(dataset_item["properties"]["datetime"])
                if not start_date.tzinfo:
                    start_date = start_date.replace(tzinfo=tzutc())
                if "completionTimeFromAscendingNode" in kwargs:
                    end_date = isoparse(kwargs["completionTimeFromAscendingNode"])
                elif "end_datetime" in dataset_item["properties"]:
                    end_date = isoparse(dataset_item["properties"]["end_datetime"])
                else:
                    end_date = today(tzinfo=tzutc())
                if not end_date.tzinfo:
                    end_date = end_date.replace(tzinfo=tzutc())

                # retrieve information about s3 from collection data
                s3_url = dataset_item["assets"]["native"]["href"]
            except KeyError as e:
                logger.warning(
                    f"Unable to extract info from {product_type} item #{i}: {str(e)}"
                )
                continue

            url_parts = urlsplit(s3_url)
            endpoint_url = url_parts.scheme + "://" + url_parts.hostname
            bucket, collection_path = get_bucket_name_and_prefix(s3_url, 0)
            if bucket is None or collection_path is None:
                logger.warning(
                    f"Unable to get bucket and prefix from {s3_url}, got {(bucket, collection_path)}"
                )
                continue

            if ".nc" in collection_path:
                num_total += 1
                if num_total < start_index:
                    continue
                if len(products) < items_per_page or items_per_page < 0:
                    product = self._create_product(
                        product_type,
                        collection_path,
                        endpoint_url + "/" + bucket,
                        dataset_item,
                        collection_dict,
                        True,
                        kwargs.get("id"),
                    )
                    if product:
                        products.append(product)
                    if product and kwargs.get("id"):
                        break
                    continue

            s3_client = _get_s3_client(endpoint_url)
            stop_search = False
            current_object = None
            while not stop_search:
                # list_objects returns max 1000 objects -> use marker to get next objects
                if current_object:
                    s3_objects = s3_client.list_objects(
                        Bucket=bucket, Prefix=collection_path, Marker=current_object
                    )
                else:
                    s3_objects = s3_client.list_objects(
                        Bucket=bucket, Prefix=collection_path
                    )
                if "Contents" not in s3_objects:
                    if len(products) == 0 and i == len(datasets_items_list) - 1:
                        return ([], 0) if prep.count else ([], None)
                    else:
                        break

                if "id" in kwargs:
                    product = self._get_product_by_id(
                        s3_objects,
                        kwargs["id"],
                        endpoint_url + "/" + bucket,
                        product_type,
                        dataset_item,
                        collection_dict,
                    )
                    if product:
                        return [product], 1
                    current_object = s3_objects["Contents"][-1]["Key"]
                    continue

                for obj in s3_objects["Contents"]:
                    item_key = obj["Key"]
                    item_id = os.path.splitext(item_key.split("/")[-1])[0]
                    # filter according to date(s) in item id
                    item_dates = re.findall(r"(\d{4})(0[1-9]|1[0-2])([0-3]\d)", item_id)
                    if not item_dates:
                        item_dates = re.findall(r"_(\d{4})(0[1-9]|1[0-2])", item_id)
                    item_dates = [
                        "".join(row) for row in item_dates
                    ]  # join tuples returned by findall
                    item_start = None
                    item_end = None
                    use_dataset_dates = False
                    if item_dates:
                        item_start = _get_date_from_yyyymmdd(item_dates[0], item_key)
                        if len(item_dates) > 2:  # start, end and created_at timestamps
                            item_end = _get_date_from_yyyymmdd(item_dates[1], item_key)
                    if not item_start:
                        # no valid datetime given in id
                        use_dataset_dates = True
                        dates = _get_dates_from_dataset_data(dataset_item)
                        if dates:
                            item_start_str = dates["start"].replace("Z", "+0000")
                            item_end_str = dates["end"].replace("Z", "+0000")
                            try:
                                item_start = datetime.strptime(
                                    item_start_str, "%Y-%m-%dT%H:%M:%S.%f%z"
                                )
                                item_end = datetime.strptime(
                                    item_end_str, "%Y-%m-%dT%H:%M:%S.%f%z"
                                )
                            except ValueError:
                                item_start = datetime.strptime(
                                    item_start_str, "%Y-%m-%dT%H:%M:%S%z"
                                )
                                item_end = datetime.strptime(
                                    item_end_str, "%Y-%m-%dT%H:%M:%S%z"
                                )
                    if not item_start:
                        # no valid datetime in id and dataset data
                        continue
                    if item_start > end_date:
                        stop_search = True
                    if (
                        (start_date <= item_start <= end_date)
                        or (item_end and start_date <= item_end <= end_date)
                        or (
                            item_end and item_start < start_date and item_end > end_date
                        )
                    ):
                        num_total += 1
                        if num_total < start_index:
                            continue
                        if len(products) < items_per_page or items_per_page < 0:
                            product = self._create_product(
                                product_type,
                                item_key,
                                endpoint_url + "/" + bucket,
                                dataset_item,
                                collection_dict,
                                use_dataset_dates,
                            )
                            if product:
                                products.append(product)
                    current_object = item_key

        return products, num_total



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from typing import Any, Optional


def check_ipython() -> bool:
    """Check if called from ipython"""
    try:
        __IPYTHON__  # type: ignore[name-defined]
        return True
    except NameError:
        return False


def check_notebook() -> bool:
    """Check if called from a notebook"""
    try:
        shell = get_ipython().__class__.__name__  # type: ignore[name-defined]
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type
    except NameError:
        return False  # Probably standard Python interpreter


class NotebookWidgets:
    """Display / handle ipython widgets"""

    is_notebook: bool = False
    html_box: Optional[Any] = None
    html_box_shown: bool = False
    display: Any = None

    def __init__(self) -> None:
        self.is_notebook = check_notebook()

        if self.is_notebook:
            from IPython.display import HTML, display, update_display

            self.display = display
            self._update_display = update_display
            self.html_box = HTML()
        else:
            pass

    def display_html(self, html_value: str) -> None:
        """Display HTML message"""

        if not self.is_notebook:
            return None

        setattr(self.html_box, "data", html_value)

        if not self.html_box_shown:
            self._html_handle = self.display(self.html_box, display_id=True)
            self.html_box_shown = True
        else:
            self._update_display(self.html_box, display_id=self._html_handle.display_id)

    def clear_html(self) -> None:
        """Clear HTML message"""

        if not self.is_notebook:
            return None

        setattr(self.html_box, "data", "")
        self._update_display(self.html_box, display_id=self._html_handle.display_id)



# -*- coding: utf-8 -*-
# Copyright 2023, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Any, ClassVar, Optional, Union

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    SerializationInfo,
    SerializerFunctionWrapHandler,
    field_validator,
    model_serializer,
    model_validator,
)

from eodag.rest.types.eodag_search import EODAGSearch
from eodag.rest.utils.rfc3339 import str_to_interval
from eodag.types import python_field_definition_to_json

if TYPE_CHECKING:
    from pydantic.fields import FieldInfo
    from typing_extensions import Self


class QueryablesGetParams(BaseModel):
    """Store GET Queryables query params"""

    collection: Optional[str] = Field(default=None, serialization_alias="productType")
    datetime: Optional[str] = Field(default=None)
    start_datetime: Optional[str] = Field(default=None)
    end_datetime: Optional[str] = Field(default=None)

    model_config = ConfigDict(extra="allow", frozen=True)

    @model_serializer(mode="wrap")
    def _serialize(self, handler: SerializerFunctionWrapHandler) -> dict[str, Any]:
        dumped: dict[str, Any] = handler(self)
        return {EODAGSearch.to_eodag(k): v for k, v in dumped.items()}

    @field_validator("datetime", "start_datetime", "end_datetime", mode="before")
    def validate_datetime(cls, value: Any) -> Optional[str]:
        """datetime, start_datetime and end_datetime must be a string"""
        if isinstance(value, list):
            return value[0]

        return value

    @model_validator(mode="after")
    def compute_datetimes(self: Self) -> Self:
        """Start datetime must be a string"""
        if not self.datetime:
            return self

        start, end = str_to_interval(self.datetime)

        if not self.start_datetime and start:
            self.start_datetime = start.strftime("%Y-%m-%dT%H:%M:%SZ")

        if not self.end_datetime and end:
            self.end_datetime = end.strftime("%Y-%m-%dT%H:%M:%SZ")

        return self


class StacQueryableProperty(BaseModel):
    """A class representing a queryable property.

    :param description: The description of the queryables property
    :param ref: (optional) A reference link to the schema of the property.
    :param type: (optional) possible types of the property
    """

    description: str
    ref: Optional[str] = Field(default=None, serialization_alias="$ref")
    type: Optional[Union[str, list[str]]] = None
    enum: Optional[list[Any]] = None
    value: Optional[Any] = None
    min: Optional[Union[int, list[Union[int, None]]]] = None
    max: Optional[Union[int, list[Union[int, None]]]] = None
    oneOf: Optional[list[Any]] = None
    items: Optional[Any] = None

    @classmethod
    def from_python_field_definition(
        cls, id: str, python_field_definition: Annotated[Any, FieldInfo]
    ) -> StacQueryableProperty:
        """Build Model from python_field_definition"""
        def_dict = python_field_definition_to_json(python_field_definition)

        if not def_dict.get("description"):
            def_dict["description"] = def_dict.get("title") or id

        return cls(**def_dict)

    @model_serializer(mode="wrap")
    def remove_none(
        self,
        handler: SerializerFunctionWrapHandler,
        _: SerializationInfo,
    ):
        """Remove none value property fields during serialization"""
        props: dict[str, Any] = handler(self)
        return {k: v for k, v in props.items() if v is not None}


class StacQueryables(BaseModel):
    """A class representing queryable properties for the STAC API.

    :param json_schema: The URL of the JSON schema.
    :param q_id: (optional) The identifier of the queryables.
    :param q_type: The type of the object.
    :param title: The title of the queryables.
    :param description: The description of the queryables
    :param properties: A dictionary of queryable properties.
    :param additional_properties: Whether additional properties are allowed.
    """

    json_schema: str = Field(
        default="https://json-schema.org/draft/2019-09/schema",
        serialization_alias="$schema",
    )
    q_id: Optional[str] = Field(default=None, serialization_alias="$id")
    q_type: str = Field(default="object", serialization_alias="type")
    title: str = Field(default="Queryables for EODAG STAC API")
    description: str = Field(
        default="Queryable names for the EODAG STAC API Item Search filter."
    )
    default_properties: ClassVar[dict[str, StacQueryableProperty]] = {
        "collection": StacQueryableProperty(
            description="Collection",
            ref="https://schemas.stacspec.org/v1.0.0/item-spec/json-schema/item.json#/collection",
        )
    }
    possible_properties: ClassVar[dict[str, StacQueryableProperty]] = {
        "geometry": StacQueryableProperty(
            description="Geometry",
            ref="https://schemas.stacspec.org/v1.0.0/item-spec/json-schema/item.json#/geometry",
        ),
        "datetime": StacQueryableProperty(
            description="Datetime - use parameters year, month, day, time instead if available",
            ref="https://schemas.stacspec.org/v1.0.0/item-spec/json-schema/datetime.json#/properties/datetime",
        ),
        "bbox": StacQueryableProperty(
            description="BBox",
            type="array",
            oneOf=[{"minItems": 4, "maxItems": 4}, {"minItems": 6, "maxItems": 6}],
            items={"type": "number"},
        ),
    }
    properties: dict[str, Any] = Field()
    required: Optional[list[str]] = Field(None)
    additional_properties: bool = Field(
        default=True, serialization_alias="additionalProperties"
    )

    def __contains__(self, name: str) -> bool:
        return name in self.properties



import io
import os
import re
import zipfile
from unittest import TestCase
from unittest.mock import call, patch

import boto3
from concurrent.futures import ThreadPoolExecutor
from moto import mock_aws

from tests import TEST_RESOURCES_PATH
from tests.context import (
    AwsAuth,
    EOProduct,
    InvalidDataError,
    PluginConfig,
    S3FileInfo,
    StreamResponse,
    _chunks_from_s3_objects,
    _compute_file_ranges,
    _prepare_file_in_zip,
    file_position_from_s3_zip,
    list_files_in_s3_zipped_object,
    open_s3_zipped_object,
    stream_download_from_s3,
    update_assets_from_s3,
)


def make_mock_fileinfo(
    key,
    size=10,
    data_start_offset=0,
    file_start_offset=0,
    data_type="application/octet-stream",
):
    fi = S3FileInfo(key=key, size=size, data_type=data_type, bucket_name="mybucket")
    fi.rel_path = key
    fi.futures = {}
    fi.file_start_offset = file_start_offset
    fi.data_start_offset = data_start_offset
    return fi


class TestUtilsS3(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.mock_aws = mock_aws()
        cls.mock_aws.start()

        cls.s3_client = boto3.client("s3", region_name="us-east-1")
        cls.s3_client.create_bucket(Bucket="mybucket")

        # zip to s3
        cls.s3_client.upload_file(
            os.path.join(
                TEST_RESOURCES_PATH,
                "products",
                "as_archive",
                "S2A_MSIL1C_20180101T105441_N0206_R051_T31TDH_20180101T124911.zip",
            ),
            "mybucket",
            "path/to/product.zip",
        )

        # unzipped to s3
        s3_prefix = "path/to/unzipped"
        product_dir = os.path.join(
            TEST_RESOURCES_PATH,
            "products",
            "S2A_MSIL1C_20180101T105441_N0206_R051_T31TDH_20180101T124911",
        )
        for root, _, files in os.walk(product_dir):
            for file_name in files:
                file_path = os.path.join(root, file_name)
                relative_path = os.path.relpath(file_path, product_dir)
                # windows path compatibilty
                s3_key = os.path.join(s3_prefix, relative_path).replace("\\", "/")
                cls.s3_client.upload_file(file_path, "mybucket", s3_key)

        cls.s3_client.put_object(Bucket="mybucket", Key="file1.txt", Body=b"abcdef")
        cls.s3_client.put_object(Bucket="mybucket", Key="file2.txt", Body=b"ghijkl")

    @classmethod
    def tearDownClass(cls):
        cls.mock_aws.stop()

    def setUp(self):
        self.prod = EOProduct("dummy", dict(geometry="POINT (0 0)", id="foo"))
        self.auth_plugin = AwsAuth(
            "dummy",
            PluginConfig.from_mapping(
                {
                    "type": "AwsAuth",
                    "credentials": dict(
                        aws_access_key_id="foo",
                        aws_secret_access_key="bar",
                    ),
                }
            ),
        )

    def assert_stream_response(
        self,
        response,
        expected_media_type,
        expected_filename_ext=None,
        expected_filename=None,
        expected_content=None,
        expected_files=None,
    ):
        self.assertIsInstance(response, StreamResponse)
        if expected_media_type == "multipart/mixed":
            self.assertTrue(response.media_type.startswith("multipart/mixed"))
        else:
            self.assertEqual(response.media_type, expected_media_type)
        if expected_filename_ext:
            self.assertIn("content-disposition", response.headers)
            match = re.search(
                r'filename="([^"]+)"',
                response.headers["content-disposition"],
            )
            self.assertIsNotNone(match)
            filename = match.group(1)  # type: ignore
            self.assertTrue(filename.endswith(expected_filename_ext))
        elif expected_filename:
            self.assertIn("content-disposition", response.headers)
            self.assertIn(expected_filename, response.headers["content-disposition"])
        else:
            self.assertNotIn("content-disposition", response.headers)

        # --- Content checks ---
        content = b"".join(response.content)
        if expected_content is not None:
            # For single file, raw
            self.assertEqual(content, expected_content)
        elif expected_files is not None:
            if expected_media_type == "application/zip":
                # For zipped responses: check zip content
                with io.BytesIO(content) as bio:
                    with zipfile.ZipFile(bio) as zf:
                        names = set(zf.namelist())
                        self.assertEqual(names, set(expected_files.keys()))
                        for fname, fcontent in expected_files.items():
                            self.assertEqual(zf.read(fname), fcontent)
            elif expected_media_type == "multipart/mixed":
                # For multipart: check that all filenames are present in the payload
                for fname in expected_files:
                    self.assertIn(fname.encode(), content)

    def test_utils_s3_list_files_in_s3_zipped_object(self):
        """list_files_in_s3_zipped_object must list the files in a zipped object stored in S3"""
        zip_infos = list_files_in_s3_zipped_object(
            "mybucket", "path/to/product.zip", self.s3_client
        )
        self.assertListEqual(
            [z.filename for z in zip_infos],
            [
                "GRANULE/",
                "GRANULE/L1C_T31TDH_A013204_20180101T105435/",
                "GRANULE/L1C_T31TDH_A013204_20180101T105435/IMG_DATA/",
                "GRANULE/L1C_T31TDH_A013204_20180101T105435/IMG_DATA/T31TDH_20180101T105441_B01.jp2",
                "GRANULE/L1C_T31TDH_A013204_20180101T105435/MTD_TL.xml",
                "MTD_MSIL1C.xml",
            ],
        )

    def test_utils_s3_open_s3_zipped_object(self):
        """list_files_in_s3_zipped_object must list the files in a zipped object stored in S3"""
        zip_file, cd_data = open_s3_zipped_object(
            "mybucket", "path/to/product.zip", self.s3_client
        )
        self.assertIsInstance(zip_file, zipfile.ZipFile)
        self.assertIsInstance(cd_data, bytes)
        self.assertEqual(len(zip_file.filelist), 6)

    def test_utils_s3_update_assets_from_s3_zip(self):
        """update_assets_from_s3 must update the assets of a product from a zipped object stored in S3"""
        update_assets_from_s3(
            self.prod, self.auth_plugin, content_url="s3://mybucket/path/to/product.zip"
        )
        self.assertEqual(len(self.prod.assets), 3)
        expected_assets = {
            "MTD_MSIL1C.xml": {
                "title": "MTD_MSIL1C.xml",
                "roles": ["metadata"],
                "href": "zip+s3://mybucket/path/to/product.zip!MTD_MSIL1C.xml",
                "type": "application/xml",
            },
            "MTD_TL.xml": {
                "title": "MTD_TL.xml",
                "roles": ["metadata"],
                "href": "zip+s3://mybucket/path/to/product.zip!GRANULE/L1C_T31TDH_A013204_20180101T105435/MTD_TL.xml",
                "type": "application/xml",
            },
            "T31TDH_20180101T105441_B01.jp2": {
                "title": "T31TDH_20180101T105441_B01.jp2",
                "roles": ["data"],
                "href": (
                    "zip+s3://mybucket/path/to/product.zip!"
                    "GRANULE/L1C_T31TDH_A013204_20180101T105435/IMG_DATA/T31TDH_20180101T105441_B01.jp2"
                ),
                "type": "image/jp2",
            },
        }
        for asset_name, expected in expected_assets.items():
            with self.subTest(asset=asset_name):
                self.assertDictEqual(self.prod.assets[asset_name].data, expected)

    def test_utils_s3_update_assets_from_s3(self):
        """update_assets_from_s3 must update the assets of a product from a folder stored in S3"""
        update_assets_from_s3(
            self.prod, self.auth_plugin, content_url="s3://mybucket/path/to/unzipped"
        )
        self.assertEqual(len(self.prod.assets), 3)
        expected_assets = {
            "MTD_MSIL1C.xml": {
                "title": "MTD_MSIL1C.xml",
                "roles": ["metadata"],
                "href": "s3://mybucket/path/to/unzipped/MTD_MSIL1C.xml",
                "type": "application/xml",
            },
            "MTD_TL.xml": {
                "title": "MTD_TL.xml",
                "roles": ["metadata"],
                "href": "s3://mybucket/path/to/unzipped/GRANULE/L1C_T31TDH_A013204_20180101T105435/MTD_TL.xml",
                "type": "application/xml",
            },
            "T31TDH_20180101T105441_B01.jp2": {
                "title": "T31TDH_20180101T105441_B01.jp2",
                "roles": ["data"],
                "href": (
                    "s3://mybucket/path/to/unzipped/GRANULE/"
                    "L1C_T31TDH_A013204_20180101T105435/IMG_DATA/T31TDH_20180101T105441_B01.jp2"
                ),
                "type": "image/jp2",
            },
        }
        for asset_name, expected in expected_assets.items():
            with self.subTest(asset=asset_name):
                self.assertDictEqual(self.prod.assets[asset_name].data, expected)

    def test_utils_s3_file_position_from_s3_zip(self):
        # Prepare a zip with both uncompressed and compressed files
        zip_bytes = io.BytesIO()
        with zipfile.ZipFile(zip_bytes, "w", compression=zipfile.ZIP_STORED) as zf:
            zf.writestr("intro.txt", "Intro file content.")
            zf.writestr("folder/data.txt", "This is a test.")
        zip_bytes.seek(0)
        self.s3_client.put_object(
            Bucket="mybucket", Key="test_uncompressed.zip", Body=zip_bytes.read()
        )

        # Prepare a zip with a compressed file
        zip_bytes2 = io.BytesIO()
        with zipfile.ZipFile(zip_bytes2, "w", compression=zipfile.ZIP_DEFLATED) as zf:
            zf.writestr("compressed.txt", "compressed content")
        zip_bytes2.seek(0)
        self.s3_client.put_object(
            Bucket="mybucket", Key="test_compressed.zip", Body=zip_bytes2.read()
        )

        test_cases = [
            {
                "desc": "uncompressed file found",
                "bucket": "mybucket",
                "key": "test_uncompressed.zip",
                "filename": "folder/data.txt",
                "expected_start": 103,
                "expected_size": 15,
                "expect_exception": None,
            },
            {
                "desc": "file not found",
                "bucket": "mybucket",
                "key": "test_uncompressed.zip",
                "filename": "does/not/exist.txt",
                "expected_start": None,
                "expected_size": None,
                "expect_exception": FileNotFoundError,
            },
            {
                "desc": "compressed file",
                "bucket": "mybucket",
                "key": "test_compressed.zip",
                "filename": "compressed.txt",
                "expected_start": None,
                "expected_size": None,
                "expect_exception": NotImplementedError,
            },
        ]

        for case in test_cases:
            with self.subTest(msg=case["desc"]):
                if case["expect_exception"]:
                    with self.assertRaises(case["expect_exception"]) as ctx:
                        file_position_from_s3_zip(
                            case["bucket"],
                            case["key"],
                            self.s3_client,
                            case["filename"],
                        )
                    if case["expect_exception"] is NotImplementedError:
                        self.assertIn(
                            "Only uncompressed files (ZIP_STORED) are supported.",
                            str(ctx.exception),
                        )
                else:
                    start, size = file_position_from_s3_zip(
                        case["bucket"], case["key"], self.s3_client, case["filename"]
                    )
                    self.assertIsInstance(start, int)
                    self.assertIsInstance(size, int)
                    self.assertEqual(start, case["expected_start"])
                    self.assertEqual(size, case["expected_size"])

    def test_utils_s3_open_s3_zipped_object_invalid(self):
        """Test a corrupted ZIP file to ensure you raise properly on bad EOCD."""
        self.s3_client.put_object(
            Bucket="mybucket", Key="bad.zip", Body=b"This is not a zip"
        )
        from tests.context import open_s3_zipped_object

        with self.assertRaises(InvalidDataError) as e:
            open_s3_zipped_object("mybucket", "bad.zip", self.s3_client)
        self.assertIn("EOCD signature not found", str(e.exception))

    def test_utils_s3_empty_zip(self):
        """Test an empty ZIP file to ensure it returns an empty list."""
        empty_zip = io.BytesIO()
        with zipfile.ZipFile(empty_zip, "w"):
            pass
        empty_zip.seek(0)
        self.s3_client.put_object(
            Bucket="mybucket", Key="empty.zip", Body=empty_zip.read()
        )

        files = list_files_in_s3_zipped_object("mybucket", "empty.zip", self.s3_client)
        self.assertEqual(files, [])

    def test_chunks_from_s3_objects(self):
        """Test _chunks_from_s3_objects to ensure it correctly retrieves data from S3 objects."""

        data = b"ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        self.s3_client.put_object(Bucket="mybucket", Key="alpha", Body=data)

        fi = make_mock_fileinfo("alpha", len(data))
        fi.data_start_offset = 0

        with ThreadPoolExecutor(max_workers=2) as executor:
            it = _chunks_from_s3_objects(
                self.s3_client, [fi], (None, None), range_size=10, executor=executor
            )
            idx, gen = next(it)
            chunks = b"".join(gen)
            self.assertEqual(idx, 0)
            self.assertEqual(chunks, data)

    def test_chunks_from_s3_objects_chunk_count_and_worker_usage(self):
        """Test that correct number of chunks are requested and workers are used efficiently."""

        data_map = {
            "file1": b"A" * 25,  # 3 chunks (10 + 10 + 5)
            "file2": b"B" * 30,  # 3 chunks (10 + 10 + 10)
        }
        range_size = 10

        file_infos = []
        for key, body in data_map.items():
            self.s3_client.put_object(Bucket="mybucket", Key=key, Body=body)
            fi = make_mock_fileinfo(key, len(body))
            fi.data_start_offset = 0
            file_infos.append(fi)

        with patch(
            "eodag.utils.s3.fetch_range", return_value=b"X" * range_size
        ) as mock_fetch:
            with ThreadPoolExecutor(max_workers=4) as executor:
                result = _chunks_from_s3_objects(
                    self.s3_client,
                    file_infos,
                    byte_range=(None, None),
                    range_size=range_size,
                    executor=executor,
                )

                for _ in result:
                    pass

            # Each 10-byte chunk corresponds to one fetch_range call
            expected_calls = [
                call("mybucket", "file1", 0, 9, self.s3_client),
                call("mybucket", "file1", 10, 19, self.s3_client),
                call("mybucket", "file1", 20, 24, self.s3_client),
                call("mybucket", "file2", 0, 9, self.s3_client),
                call("mybucket", "file2", 10, 19, self.s3_client),
                call("mybucket", "file2", 20, 29, self.s3_client),
            ]

            self.assertEqual(mock_fetch.call_count, 6)
            mock_fetch.assert_has_calls(expected_calls, any_order=True)

    def test_prepare_file_in_zip(self):
        """Test _prepare_file_in_zip to ensure it sets the correct attributes for retrieving a file inside a ZIP."""

        zip_bytes = io.BytesIO()
        with zipfile.ZipFile(zip_bytes, "w", compression=zipfile.ZIP_STORED) as zf:
            zf.writestr("foo.txt", "hello")
        zip_bytes.seek(0)
        self.s3_client.put_object(
            Bucket="mybucket", Key="test.zip", Body=zip_bytes.read()
        )

        fi = make_mock_fileinfo("test.zip!foo.txt", 0)
        _prepare_file_in_zip(fi, self.s3_client)

        self.assertEqual(fi.key, "test.zip")
        self.assertEqual(fi.zip_filepath, "foo.txt")
        self.assertGreater(fi.data_start_offset, 0)
        self.assertEqual(fi.size, len("hello"))

    def test_compute_file_ranges(self):
        """Test _compute_file_ranges to ensure it calculates correct byte ranges for files to download."""

        range_size = 10

        test_cases = [
            {
                "desc": "Full file, no byte_range limits (start=None, end=None)",
                "file_info": make_mock_fileinfo("a", 50),
                "byte_range": (None, None),
                "expected": [
                    (i, min(i + range_size - 1, 49)) for i in range(0, 50, range_size)
                ],
            },
            {
                "desc": "Partial overlap at start: range starts inside the file",
                "file_info": make_mock_fileinfo("a", 50, 5, 100),
                "byte_range": (110, 149),
                "expected": [(15, 24), (25, 34), (35, 44), (45, 54)],
            },
            {
                "desc": "Partial overlap at end: range ends inside the file",
                "file_info": make_mock_fileinfo("a", 50, 0, 100),
                "byte_range": (90, 120),
                "expected": [(0, 9), (10, 19), (20, 20)],
            },
            {
                "desc": "Range fully inside file",
                "file_info": make_mock_fileinfo("a", 50, 0, 0),
                "byte_range": (10, 29),
                "expected": [(10, 19), (20, 29)],
            },
            {
                "desc": "Range fully outside before file - should return None",
                "file_info": make_mock_fileinfo("a", 50, 0, 100),
                "byte_range": (0, 50),
                "expected": None,
            },
            {
                "desc": "Range fully outside after file - should return None",
                "file_info": make_mock_fileinfo("a", 50, 0, 100),
                "byte_range": (151, 200),
                "expected": None,
            },
            {
                "desc": "Open ended start range (None as start)",
                "file_info": make_mock_fileinfo("a", 50, 2, 100),
                "byte_range": (None, 105),
                "expected": [(2, 7)],
            },
            {
                "desc": "Open ended end range (None as end)",
                "file_info": make_mock_fileinfo("a", 50, 2, 100),
                "byte_range": (102, None),
                "expected": [(4, 13), (14, 23), (24, 33), (34, 43), (44, 51)],
            },
        ]

        for case in test_cases:
            with self.subTest(msg=case["desc"]):
                result = _compute_file_ranges(
                    case["file_info"], case["byte_range"], range_size
                )
                self.assertEqual(result, case["expected"])

    def test_stream_download_from_s3(self):
        test_cases = [
            {
                "desc": "single file, raw",
                "files_info": [
                    make_mock_fileinfo("file1.txt", size=6, data_type="text/plain")
                ],
                "compress": "raw",
                "zip_filename": "archive",
                "expected_media_type": "text/plain",
                "expected_filename_ext": None,
                "expected_filename": "file1.txt",
                "expected_content": b"abcdef",
            },
            {
                "desc": "multiple files, zip",
                "files_info": [
                    make_mock_fileinfo("file1.txt", size=6),
                    make_mock_fileinfo("file2.txt", size=6),
                ],
                "compress": "zip",
                "zip_filename": "myarchive",
                "expected_media_type": "application/zip",
                "expected_filename_ext": ".zip",
                "expected_files": {
                    "file1.txt": b"abcdef",
                    "file2.txt": b"ghijkl",
                },
            },
            {
                "desc": "multiple files, multipart",
                "files_info": [
                    make_mock_fileinfo("file1.txt", size=6, data_type="text/plain"),
                    make_mock_fileinfo("file2.txt", size=6, data_type="text/plain"),
                ],
                "compress": "raw",
                "zip_filename": "archive",
                "expected_media_type": "multipart/mixed",
                "expected_filename_ext": None,
                "expected_filename": None,
                "expected_files": {
                    "file1.txt": b"abcdef",
                    "file2.txt": b"ghijkl",
                },
            },
            {
                "desc": "single file, zipped",
                "files_info": [
                    make_mock_fileinfo("file1.txt", size=6, data_type="text/plain")
                ],
                "compress": "zip",
                "zip_filename": "singlefile",
                "expected_media_type": "application/zip",
                "expected_filename_ext": ".zip",
                "expected_files": {
                    "file1.txt": b"abcdef",
                },
            },
        ]

        for case in test_cases:
            with self.subTest(msg=case["desc"]):
                response = stream_download_from_s3(
                    s3_client=self.s3_client,
                    files_info=case["files_info"],
                    compress=case["compress"],
                    zip_filename=case["zip_filename"],
                    max_workers=2,
                )

                self.assert_stream_response(
                    response,
                    case["expected_media_type"],
                    expected_filename_ext=case.get("expected_filename_ext"),
                    expected_filename=case.get("expected_filename"),
                    expected_content=case.get("expected_content"),
                    expected_files=case.get("expected_files"),
                )



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging
import re
from operator import attrgetter
from typing import TYPE_CHECKING, Any, Iterator, Optional, Union, cast

import importlib_metadata

from eodag.config import (
    AUTH_TOPIC_KEYS,
    PLUGINS_TOPICS_KEYS,
    load_config,
    merge_configs,
)
from eodag.plugins.apis.base import Api
from eodag.plugins.authentication.base import Authentication
from eodag.plugins.base import EODAGPluginMount
from eodag.plugins.crunch.base import Crunch
from eodag.plugins.download.base import Download
from eodag.plugins.search.base import Search
from eodag.utils import GENERIC_PRODUCT_TYPE, deepcopy, dict_md5sum
from eodag.utils.exceptions import (
    AuthenticationError,
    MisconfiguredError,
    UnsupportedProvider,
)

if TYPE_CHECKING:
    from mypy_boto3_s3 import S3ServiceResource
    from requests.auth import AuthBase

    from eodag.api.product import EOProduct
    from eodag.config import PluginConfig, ProviderConfig
    from eodag.plugins.base import PluginTopic
    from eodag.types import S3SessionKwargs


logger = logging.getLogger("eodag.plugins.manager")


class PluginManager:
    """A manager for the plugins.

    The role of instances of this class (normally only one instance exists,
    created during instantiation of :class:`~eodag.api.core.EODataAccessGateway`.
    But nothing is done to enforce this) is to instantiate the plugins
    according to the providers configuration, keep track of them in memory, and
    manage a cache of plugins already constructed. The providers configuration contains
    information such as the name of the provider, the internet endpoint for accessing
    it, and the plugins to use to perform defined actions (search, download,
    authenticate, crunch).

    :param providers_config: The configuration with all information about the providers
                             supported by ``eodag``
    """

    supported_topics = set(PLUGINS_TOPICS_KEYS)

    product_type_to_provider_config_map: dict[str, list[ProviderConfig]]

    skipped_plugins: list[str]

    def __init__(self, providers_config: dict[str, ProviderConfig]) -> None:
        self.skipped_plugins = []
        self.providers_config = providers_config
        # Load all the plugins. This will make all plugin classes of a particular
        # type to be available in the base plugin class's 'plugins' attribute.
        # For example, by importing module 'eodag.plugins.search.resto', the plugin
        # 'RestoSearch' will be available in self.supported_topics['search'].plugins
        for topic in self.supported_topics:
            # This way of discovering plugins means that anyone can create eodag
            # plugins as a separate python package (though it must require eodag), and
            # have it discovered as long as they declare an entry point of the type
            # 'eodag.plugins.search' for example in its setup script. See the setup
            # script of eodag for an example of how to do this.
            for entry_point in importlib_metadata.entry_points(
                group="eodag.plugins.{}".format(topic)
            ):
                try:
                    entry_point.load()
                except ModuleNotFoundError:
                    logger.debug(
                        "%s plugin skipped, eodag[%s] or eodag[all] needed",
                        entry_point.name,
                        ",".join(entry_point.extras),
                    )
                    self.skipped_plugins.append(entry_point.name)
                except ImportError:
                    import traceback as tb

                    logger.warning("Unable to load plugin: %s.", entry_point.name)
                    logger.warning("Reason:\n%s", tb.format_exc())
                    logger.warning(
                        "Check that the plugin module (%s) is importable",
                        entry_point.name,
                    )
                if entry_point.dist and entry_point.dist.name != "eodag":
                    # use plugin providers if any
                    name = entry_point.dist.name
                    dist = entry_point.dist
                    plugin_providers_config_path = [
                        str(x) for x in dist.locate_file(name).rglob("providers.yml")
                    ]
                    if plugin_providers_config_path:
                        plugin_providers_config = load_config(
                            plugin_providers_config_path[0]
                        )
                        merge_configs(plugin_providers_config, self.providers_config)
                        self.providers_config = plugin_providers_config
        self.rebuild()

    def rebuild(
        self, providers_config: Optional[dict[str, ProviderConfig]] = None
    ) -> None:
        """(Re)Build plugin manager mapping and cache"""
        if providers_config is not None:
            self.providers_config = providers_config

        self.build_product_type_to_provider_config_map()
        self._built_plugins_cache: dict[tuple[str, str, str], Any] = {}

    def build_product_type_to_provider_config_map(self) -> None:
        """Build mapping conf between product types and providers"""
        self.product_type_to_provider_config_map = {}
        for provider in list(self.providers_config):
            provider_config = self.providers_config[provider]
            if not hasattr(provider_config, "products") or not provider_config.products:
                logger.info(
                    "%s: provider has no product configured and will be skipped",
                    provider,
                )
                self.providers_config.pop(provider)
                continue

            # provider priority set to lowest if not set
            if getattr(provider_config, "priority", None) is None:
                self.providers_config[provider].priority = provider_config.priority = 0

            for product_type in provider_config.products:
                product_type_providers = (
                    self.product_type_to_provider_config_map.setdefault(  # noqa
                        product_type, []
                    )
                )
                product_type_providers.append(provider_config)
                product_type_providers.sort(key=attrgetter("priority"), reverse=True)

    def get_search_plugins(
        self, product_type: Optional[str] = None, provider: Optional[str] = None
    ) -> Iterator[Union[Search, Api]]:
        """Build and return all the search plugins supporting the given product type,
        ordered by highest priority, or the search plugin of the given provider

        :param product_type: (optional) The product type that the constructed plugins
                             must support
        :param provider: (optional) The provider or the provider group on which to get
            the search plugins
        :returns: All the plugins supporting the product type, one by one (a generator
                  object)
            or :class:`~eodag.plugins.download.Api`)
        :raises: :class:`~eodag.utils.exceptions.UnsupportedProvider`
        """

        def get_plugin() -> Union[Search, Api]:
            plugin: Union[Search, Api]
            if search := getattr(config, "search", None):
                config.search.products = config.products
                config.search.priority = config.priority
                plugin = cast(Search, self._build_plugin(config.name, search, Search))
            elif api := getattr(config, "api", None):
                config.api.products = config.products
                config.api.priority = config.priority
                plugin = cast(Api, self._build_plugin(config.name, api, Api))
            else:
                raise MisconfiguredError(
                    f"No search plugin configured for {config.name}."
                )
            return plugin

        configs: Optional[list[ProviderConfig]]
        if product_type:
            configs = self.product_type_to_provider_config_map.get(product_type)
            if not configs:
                logger.info(
                    "UnsupportedProductType: %s, using generic settings", product_type
                )
                configs = self.product_type_to_provider_config_map[GENERIC_PRODUCT_TYPE]
        else:
            configs = list(self.providers_config.values())

        if provider:
            configs = [
                c for c in configs if provider in [getattr(c, "group", None), c.name]
            ]

        if not configs and product_type:
            raise UnsupportedProvider(
                f"{provider} is not (yet) supported for {product_type}"
            )
        if not configs:
            raise UnsupportedProvider(f"{provider} is not (yet) supported")

        for config in sorted(configs, key=attrgetter("priority"), reverse=True):
            yield get_plugin()

    def get_download_plugin(self, product: EOProduct) -> Union[Download, Api]:
        """Build and return the download plugin capable of downloading the given
        product.

        :param product: The product to get a download plugin for
        :returns: The download plugin capable of downloading the product
        """
        plugin_conf = self.providers_config[product.provider]
        if download := getattr(plugin_conf, "download", None):
            plugin_conf.download.priority = plugin_conf.priority
            plugin = cast(
                Download,
                self._build_plugin(product.provider, download, Download),
            )
        elif api := getattr(plugin_conf, "api", None):
            plugin_conf.api.products = plugin_conf.products
            plugin_conf.api.priority = plugin_conf.priority
            plugin = cast(Api, self._build_plugin(product.provider, api, Api))
        else:
            raise MisconfiguredError(
                f"No download plugin configured for provider {plugin_conf.name}."
            )
        return plugin

    def get_auth_plugin(
        self, associated_plugin: PluginTopic, product: Optional[EOProduct] = None
    ) -> Optional[Authentication]:
        """Build and return the authentication plugin associated to the given
        search/download plugin

        .. versionchanged:: v3.0.0
            ``get_auth_plugin()`` now needs ``associated_plugin`` instead of ``provider``
            as argument.

        :param associated_plugin: The search/download plugin to which the authentication
                                  plugin is linked
        :param product: The product to download. ``None`` for search authentication
        :returns: The Authentication plugin
        """
        # matching url from product to download
        if product is not None and len(product.assets) > 0:
            matching_url = next(iter(product.assets.values()))["href"]
        elif product is not None:
            matching_url = product.properties.get(
                "downloadLink"
            ) or product.properties.get("orderLink")
        else:
            # search auth
            matching_url = getattr(associated_plugin.config, "api_endpoint", None)

        try:
            auth_plugin = next(
                self.get_auth_plugins(
                    associated_plugin.provider,
                    matching_url=matching_url,
                    matching_conf=associated_plugin.config,
                )
            )
        except StopIteration:
            auth_plugin = None
        return auth_plugin

    def get_auth_plugins(
        self,
        provider: str,
        matching_url: Optional[str] = None,
        matching_conf: Optional[PluginConfig] = None,
    ) -> Iterator[Authentication]:
        """Build and return the authentication plugin for the given product_type and
        provider

        :param provider: The provider for which to get the authentication plugin
        :param matching_url: url to compare with plugin matching_url pattern
        :param matching_conf: configuration to compare with plugin matching_conf
        :returns: All the Authentication plugins for the given criteria
        """
        auth_conf: Optional[PluginConfig] = None

        def _is_auth_plugin_matching(
            auth_conf: PluginConfig,
            matching_url: Optional[str],
            matching_conf: Optional[PluginConfig],
        ) -> bool:
            plugin_matching_conf = getattr(auth_conf, "matching_conf", {})
            if matching_conf:
                if (
                    plugin_matching_conf
                    and matching_conf.__dict__.items() >= plugin_matching_conf.items()
                ):
                    # conf matches
                    return True
            plugin_matching_url = getattr(auth_conf, "matching_url", None)
            if matching_url:
                if plugin_matching_url and re.match(
                    rf"{plugin_matching_url}", matching_url
                ):
                    # url matches
                    return True
            # no match
            return False

        # providers configs with given provider at first
        sorted_providers_config = deepcopy(self.providers_config)
        sorted_providers_config = {
            provider: sorted_providers_config.pop(provider),
            **sorted_providers_config,
        }

        for plugin_provider, provider_conf in sorted_providers_config.items():
            for key in AUTH_TOPIC_KEYS:
                auth_conf = getattr(provider_conf, key, None)
                if auth_conf is None:
                    continue
                # plugin without configured match criteria: only works for given provider
                unconfigured_match = (
                    True
                    if (
                        not getattr(auth_conf, "matching_conf", {})
                        and not getattr(auth_conf, "matching_url", None)
                        and provider == plugin_provider
                    )
                    else False
                )

                if unconfigured_match or _is_auth_plugin_matching(
                    auth_conf, matching_url, matching_conf
                ):
                    auth_conf.priority = provider_conf.priority
                    plugin = cast(
                        Authentication,
                        self._build_plugin(plugin_provider, auth_conf, Authentication),
                    )
                    yield plugin
                else:
                    continue

    def get_auth(
        self,
        provider: str,
        matching_url: Optional[str] = None,
        matching_conf: Optional[PluginConfig] = None,
    ) -> Optional[Union[AuthBase, S3SessionKwargs, S3ServiceResource]]:
        """Authenticate and return the authenticated object for the first matching
        authentication plugin

        :param provider: The provider for which to get the authentication plugin
        :param matching_url: url to compare with plugin matching_url pattern
        :param matching_conf: configuration to compare with plugin matching_conf
        :returns: All the Authentication plugins for the given criteria
        """
        for auth_plugin in self.get_auth_plugins(provider, matching_url, matching_conf):
            if auth_plugin and callable(getattr(auth_plugin, "authenticate", None)):
                try:
                    auth = auth_plugin.authenticate()
                    return auth
                except (AuthenticationError, MisconfiguredError) as e:
                    logger.debug(f"Could not authenticate on {provider}: {str(e)}")
                    continue
            else:
                logger.debug(
                    f"Could not authenticate on {provider} using {auth_plugin} plugin"
                )
                continue
        return None

    @staticmethod
    def get_crunch_plugin(name: str, **options: Any) -> Crunch:
        """Instantiate a eodag Crunch plugin whom class name is `name`, and configure
        it with the `options`

        :param name: The name of the Crunch plugin to instantiate
        :param options: The configuration parameters of the cruncher
        :returns: The cruncher named `name`
        """
        klass = Crunch.get_plugin_by_class_name(name)
        return klass(options)

    def sort_providers(self) -> None:
        """Sort providers taking into account current priority order"""
        for provider_configs in self.product_type_to_provider_config_map.values():
            provider_configs.sort(key=attrgetter("priority"), reverse=True)

    def set_priority(self, provider: str, priority: int) -> None:
        """Set the priority of the given provider

        :param provider: The provider which is assigned the priority
        :param priority: The priority to assign to the provider
        """
        # Update the priority in the configurations so that it is taken into account
        # when a plugin of this provider is latterly built
        for (
            _,
            provider_configs,
        ) in self.product_type_to_provider_config_map.items():
            for config in provider_configs:
                if config.name == provider:
                    config.priority = priority
            # Sort the provider configs, taking into account the new priority order
            provider_configs.sort(key=attrgetter("priority"), reverse=True)
        # Update the priority of already built plugins of the given provider
        for provider_name, topic_class, auth_match_md5 in self._built_plugins_cache:
            if provider_name == provider:
                self._built_plugins_cache[
                    (provider, topic_class, auth_match_md5)
                ].priority = priority

    def _build_plugin(
        self,
        provider: str,
        plugin_conf: PluginConfig,
        topic_class: type[PluginTopic],
    ) -> Union[Api, Search, Download, Authentication, Crunch]:
        """Build the plugin of the given topic with the given plugin configuration and
        registered as the given provider

        :param provider: The provider for which to build the plugin
        :param plugin_conf: The configuration of the plugin to be built
        :param topic_class: The type of plugin to build
        :returns: The built plugin
                :class:`~eodag.plugin.download.Download` or
                :class:`~eodag.plugin.authentication.Authentication` or
                :class:`~eodag.plugin.crunch.Crunch`
        """
        # md5 hash to helps identifying an auth plugin within several for a given provider
        # (each has distinct matching settings)
        auth_match_md5 = dict_md5sum(
            {
                "matching_url": getattr(plugin_conf, "matching_url", None),
                "matching_conf": getattr(plugin_conf, "matching_conf", None),
            }
        )
        cached_instance = self._built_plugins_cache.setdefault(
            (provider, topic_class.__name__, auth_match_md5), None
        )
        if cached_instance is not None:
            return cached_instance
        plugin_class = EODAGPluginMount.get_plugin_by_class_name(
            topic_class, getattr(plugin_conf, "type")
        )
        plugin: Union[Api, Search, Download, Authentication, Crunch] = plugin_class(
            provider, plugin_conf
        )
        self._built_plugins_cache[
            (provider, topic_class.__name__, auth_match_md5)
        ] = plugin
        return plugin



# -*- coding: utf-8 -*-
# Copyright 2024, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from types import MethodType
from typing import Any

from botocore.exceptions import BotoCoreError

from eodag.api.product import EOProduct  # type: ignore
from eodag.api.search_result import RawSearchResult
from eodag.plugins.search.qssearch import ODataV4Search
from eodag.utils.exceptions import MisconfiguredError, RequestError
from eodag.utils.s3 import update_assets_from_s3

logger = logging.getLogger("eodag.search.creodiass3")


def patched_register_downloader(self, downloader, authenticator):
    """Add the download information to the product.

    :param self: product to which information should be added
    :param downloader: The download method that it can use
                      :class:`~eodag.plugins.download.base.Download` or
                      :class:`~eodag.plugins.api.base.Api`
    :param authenticator: The authentication method needed to perform the download
                         :class:`~eodag.plugins.authentication.base.Authentication`
    """
    # verify credentials
    required_creds = ["aws_access_key_id", "aws_secret_access_key"]
    credentials = getattr(authenticator.config, "credentials", {}) or {}
    if not all(x in credentials and credentials[x] for x in required_creds):
        raise MisconfiguredError(
            f"Incomplete credentials for {self.provider}, missing "
            f"{[x for x in required_creds if x not in credentials or not credentials[x]]}"
        )
    # register downloader
    self.register_downloader_only(downloader, authenticator)
    # and also update assets
    try:
        update_assets_from_s3(
            self, authenticator, getattr(downloader.config, "s3_endpoint", None)
        )
    except BotoCoreError as e:
        raise RequestError.from_error(e, "could not update assets") from e


class CreodiasS3Search(ODataV4Search):
    """
    ``CreodiasS3Search`` is an extension of :class:`~eodag.plugins.search.qssearch.ODataV4Search`,
    it executes a Search on creodias and adapts results so that the assets contain links to s3.
    It has the same configuration parameters as :class:`~eodag.plugins.search.qssearch.ODataV4Search` and
    one additional parameter:

    :param provider: provider name
    :param config: Search plugin configuration:

        * :attr:`~eodag.config.PluginConfig.s3_endpoint` (``str``) (**mandatory**): base url of the s3
    """

    def __init__(self, provider, config):
        super(CreodiasS3Search, self).__init__(provider, config)

    def normalize_results(
        self, results: RawSearchResult, **kwargs: Any
    ) -> list[EOProduct]:
        """Build EOProducts from provider results"""

        products = super(CreodiasS3Search, self).normalize_results(results, **kwargs)

        for product in products:
            # backup original register_downloader to register_downloader_only
            product.register_downloader_only = product.register_downloader
            # patched register_downloader that will also update assets
            product.register_downloader = MethodType(
                patched_register_downloader, product
            )

        return products



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from datetime import datetime
from typing import Annotated, Any, Optional, Union, cast

from annotated_types import MinLen
from pydantic import BaseModel, ConfigDict, Field, conint, field_validator
from shapely import wkt
from shapely.errors import GEOSException
from shapely.geometry import Polygon, shape
from shapely.geometry.base import GEOMETRY_TYPES, BaseGeometry

from eodag.types.bbox import BBox
from eodag.utils import DEFAULT_ITEMS_PER_PAGE, DEFAULT_PAGE
from eodag.utils.exceptions import ValidationError

NumType = Union[float, int]
GeomArgs = Union[list[NumType], tuple[NumType], dict[str, NumType], str, BaseGeometry]

PositiveInt = conint(gt=0)
SortByList = Annotated[list[tuple[str, str]], MinLen(1)]


class SearchArgs(BaseModel):
    """Represents an EODAG search"""

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)

    provider: Optional[str] = Field(None)
    productType: str = Field()
    id: Optional[str] = Field(None)
    start: Optional[str] = Field(None)
    end: Optional[str] = Field(None)
    geom: Optional[BaseGeometry] = Field(None)
    locations: Optional[dict[str, str]] = Field(None)
    page: Optional[int] = Field(DEFAULT_PAGE, gt=0)  # type: ignore
    items_per_page: Optional[PositiveInt] = Field(DEFAULT_ITEMS_PER_PAGE)  # type: ignore
    sort_by: Optional[SortByList] = Field(None)  # type: ignore

    @field_validator("start", "end", mode="before")
    @classmethod
    def check_date_format(cls, v: str) -> str:
        """Validate dates"""
        try:
            datetime.fromisoformat(v)
        except ValueError as e:
            raise ValueError("start_date and end must follow ISO8601 format") from e
        return v

    @field_validator("geom", mode="before")
    @classmethod
    def check_geom(cls, v: Any) -> BaseGeometry:
        """Validate geom"""
        # GeoJSON geometry
        if isinstance(v, dict) and v.get("type") in GEOMETRY_TYPES:  # type: ignore
            return cast(BaseGeometry, shape(v))

        # Bounding Box
        if isinstance(v, (list, tuple, dict)):
            return BBox(v).to_polygon()  # type: ignore

        if isinstance(v, str):
            # WKT geometry
            try:
                return cast(Polygon, wkt.loads(v))  # type: ignore
            except GEOSException as e:  # type: ignore
                raise ValueError(f"Invalid geometry WKT string: {v}") from e

        if isinstance(v, BaseGeometry):
            return v

        raise TypeError(f"Invalid geometry type: {type(v)}")

    @field_validator("sort_by", mode="before")
    @classmethod
    def check_sort_by_arg(
        cls,
        sort_by_arg: Optional[SortByList],  # type: ignore
    ) -> Optional[SortByList]:  # type: ignore
        """Check if the sort_by argument is correct

        :param sort_by_arg: The sort_by argument
        :returns: The sort_by argument with sorting order parsed (whitespace(s) are
                  removed and only the 3 first letters in uppercase are kept)
        """
        if sort_by_arg is None:
            return None

        assert isinstance(
            sort_by_arg, list
        ), f"Sort argument must be a list of tuple(s), got a '{type(sort_by_arg)}' instead"
        sort_order_pattern = r"^(ASC|DES)[a-zA-Z]*$"
        for i, sort_by_tuple in enumerate(sort_by_arg):
            assert isinstance(
                sort_by_tuple, tuple
            ), f"Sort argument must be a list of tuple(s), got a list of '{type(sort_by_tuple)}' instead"
            # get sorting elements by removing leading and trailing whitespace(s) if exist
            sort_param = sort_by_tuple[0].strip()
            sort_order = sort_by_tuple[1].strip().upper()
            assert re.match(sort_order_pattern, sort_order) is not None, (
                "Sorting order must be set to 'ASC' (ASCENDING) or 'DESC' (DESCENDING), "
                f"got '{sort_order}' with '{sort_param}' instead"
            )
            sort_by_arg[i] = (sort_param, sort_order[:3])
        # remove duplicates
        pruned_sort_by_arg: SortByList = list(dict.fromkeys(sort_by_arg))  # type: ignore
        for i, sort_by_tuple in enumerate(pruned_sort_by_arg):
            for j, sort_by_tuple_tmp in enumerate(pruned_sort_by_arg):
                # since duplicated tuples or dictionnaries have been removed, if two sorting parameters are equal,
                # then their sorting order is different and there is a contradiction that would raise an error
                if i != j and sort_by_tuple[0] == sort_by_tuple_tmp[0]:
                    raise ValidationError(
                        f"'{sort_by_tuple[0]}' parameter is called several times to sort results with different "
                        "sorting orders. Please set it to only one ('ASC' (ASCENDING) or 'DESC' (DESCENDING))",
                        set([sort_by_tuple[0]]),
                    )
        return pruned_sort_by_arg



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Annotated, Any, Optional
from unittest import mock

import geojson
from pydantic.fields import FieldInfo

from eodag.api.product.metadata_mapping import get_metadata_path_value
from eodag.api.search_result import SearchResult
from eodag.plugins.crunch.filter_date import FilterDate
from eodag.plugins.crunch.filter_overlap import FilterOverlap
from eodag.plugins.crunch.filter_property import FilterProperty
from eodag.plugins.search import PreparedSearch
from eodag.plugins.search.qssearch import StacSearch
from eodag.types.queryables import Queryables
from eodag.utils import HTTP_REQ_TIMEOUT, MockResponse
from eodag.utils.stac_reader import fetch_stac_collections, fetch_stac_items

if TYPE_CHECKING:
    from eodag.api.product import EOProduct
    from eodag.config import PluginConfig


logger = logging.getLogger("eodag.search.static_stac_search")


class StaticStacSearch(StacSearch):
    """Static STAC Catalog search plugin

    This plugin first loads all STAC items found in the catalog (or item), and converts them to
    EOProducts using :class:`~eodag.plugins.search.qssearch.StacSearch`.
    Then it uses crunchers to only keep products matching query parameters.

    The plugin inherits the configuration parameters from :class:`~eodag.plugins.search.qssearch.PostJsonSearch`
    (via the :class:`~eodag.plugins.search.qssearch.StacSearch` inheritance) with the following particularities:

    :param provider: provider name
    :param config: Search plugin configuration:

        * :attr:`~eodag.config.PluginConfig.api_endpoint` (``str``) (**mandatory**): path to the catalog or item;
          in contrast to the api_endpoint for other plugin types this can be a url or local system path.
        * :attr:`~eodag.config.PluginConfig.max_connections` (``int``): Maximum number of concurrent
          connections for HTTP requests; default: ``100``
        * :attr:`~eodag.config.PluginConfig.timeout` (``int``): Timeout in seconds for each
          internal HTTP request; default: ``5``

    """

    def __init__(self, provider: str, config: PluginConfig) -> None:
        # prevent search parameters from being queried when they are known in the configuration or not
        for param, mapping in config.metadata_mapping.items():
            # only keep one queryable to allow the mock search request
            if param != "productType":
                config.metadata_mapping[param] = get_metadata_path_value(mapping)
        config.discover_metadata["auto_discovery"] = False
        # there is no endpoint for fetching queryables with a static search
        config.discover_queryables["fetch_url"] = None
        config.discover_queryables["product_type_fetch_url"] = None

        super(StaticStacSearch, self).__init__(provider, config)
        self.config.__dict__.setdefault("max_connections", 100)
        self.config.__dict__.setdefault("timeout", HTTP_REQ_TIMEOUT)
        self.config.__dict__.setdefault("ssl_verify", True)
        self.config.__dict__.setdefault("pagination", {})
        self.config.__dict__["pagination"].setdefault(
            "total_items_nb_key_path", "$.null"
        )
        self.config.__dict__["pagination"].setdefault("max_items_per_page", -1)
        # disable product types discovery by default (if endpoints equals to STAC API default)
        if (
            getattr(self.config, "discover_product_types", {}).get("fetch_url")
            == "{api_endpoint}/../collections"
        ):
            self.config.discover_product_types = {}

    def discover_product_types(self, **kwargs: Any) -> Optional[dict[str, Any]]:
        """Fetch product types list from a static STAC Catalog provider using `discover_product_types` conf

        :returns: configuration dict containing fetched product types information
        """
        unformatted_fetch_url = self.config.discover_product_types.get("fetch_url")
        if unformatted_fetch_url is None:
            return None
        fetch_url = unformatted_fetch_url.format(**self.config.__dict__)

        collections = fetch_stac_collections(
            fetch_url,
            collection=kwargs.get("q"),
            max_connections=self.config.max_connections,
            timeout=int(self.config.timeout),
            ssl_verify=self.config.ssl_verify,
        )
        if "q" in kwargs:
            collections = [c for c in collections if c["id"] == kwargs["q"]]
        collections_mock_response = {"collections": collections}

        # discover_product_types on mocked QueryStringSearch._request
        with mock.patch(
            "eodag.plugins.search.qssearch.QueryStringSearch._request",
            autospec=True,
            return_value=MockResponse(collections_mock_response, 200),
        ):
            conf_update_dict = super(StaticStacSearch, self).discover_product_types(
                **kwargs
            )

        return conf_update_dict

    def discover_queryables(
        self, **kwargs: Any
    ) -> dict[str, Annotated[Any, FieldInfo]]:
        """Set static available queryables for :class:`~eodag.plugins.search.static_stac_search.StaticStacSearch`
        search plugin

        :param kwargs: additional filters for queryables (`productType` and other search
                       arguments)
        :returns: queryable parameters dict
        """
        return {
            "productType": Queryables.get_with_default(
                "productType", kwargs.get("productType")
            ),
            "id": Queryables.get_with_default("id", kwargs.get("id")),
            "start": Queryables.get_with_default(
                "start", kwargs.get("start") or kwargs.get("startTimeFromAscendingNode")
            ),
            "end": Queryables.get_with_default(
                "end",
                kwargs.get("end") or kwargs.get("completionTimeFromAscendingNode"),
            ),
            "geom": Queryables.get_with_default(
                "geom",
                kwargs.get("geom") or kwargs.get("geometry") or kwargs.get("area"),
            ),
        }

    def query(
        self,
        prep: PreparedSearch = PreparedSearch(),
        **kwargs: Any,
    ) -> tuple[list[EOProduct], Optional[int]]:
        """Perform a search on a static STAC Catalog"""

        # only return 1 page if pagination is disabled
        if (
            prep.page
            and prep.page > 1
            and prep.items_per_page is not None
            and prep.items_per_page <= 0
        ):
            return ([], 0) if prep.count else ([], None)

        product_type = kwargs.get("productType", prep.product_type)
        # provider product type specific conf
        self.product_type_def_params = (
            self.get_product_type_def_params(product_type, format_variables=kwargs)
            if product_type is not None
            else {}
        )

        for collection in self.get_collections(prep, **kwargs):
            # skip empty collection if one is required in api_endpoint
            if "{collection}" in self.config.api_endpoint and not collection:
                continue
            search_endpoint = self.config.api_endpoint.rstrip("/").format(
                collection=collection
            )

        features = fetch_stac_items(
            search_endpoint,
            recursive=True,
            max_connections=self.config.max_connections,
            timeout=int(self.config.timeout),
            ssl_verify=self.config.ssl_verify,
        )
        nb_features = len(features)
        feature_collection = geojson.FeatureCollection(features)

        # query on mocked StacSearch._request
        with mock.patch(
            "eodag.plugins.search.qssearch.StacSearch._request",
            autospec=True,
            return_value=MockResponse(feature_collection, 200),
        ):
            eo_products, _ = super(StaticStacSearch, self).query(
                PreparedSearch(items_per_page=nb_features, page=1, count=True), **kwargs
            )
        # filter using query params
        search_result = SearchResult(eo_products)
        # Filter by date
        if "startTimeFromAscendingNode" in kwargs:
            kwargs["start"] = kwargs.pop("startTimeFromAscendingNode")
        if "completionTimeFromAscendingNode" in kwargs:
            kwargs["end"] = kwargs.pop("completionTimeFromAscendingNode")
        if any(k in ["start", "end"] for k in kwargs.keys()):
            search_result = search_result.crunch(
                FilterDate({k: kwargs[k] for k in ["start", "end"] if k in kwargs})
            )

        # Filter by geometry
        geometry = kwargs.pop("geometry", None)
        if geometry:
            search_result = search_result.crunch(
                FilterOverlap({"intersects": True}), geometry=geometry
            )
        # Filter by cloudCover
        if "cloudCover" in kwargs.keys():
            search_result = search_result.crunch(
                FilterProperty(
                    {"cloudCover": kwargs.pop("cloudCover"), "operator": "lt"}
                )
            )
        # Filter by other properties
        skip_eodag_internal_parameters = [
            "auth",
            "raise_errors",
            "productType",
            "locations",
            "start",
            "end",
            "geom",
        ]
        for property_key, property_value in kwargs.items():
            if property_key not in skip_eodag_internal_parameters:
                search_result = search_result.crunch(
                    FilterProperty({property_key: property_value, "operator": "eq"})
                )

        return (
            (search_result.data, len(search_result))
            if prep.count
            else (search_result.data, None)
        )



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import datetime
import logging
import time
from datetime import datetime as dt
from typing import TYPE_CHECKING, Any

import dateutil.parser
from dateutil import tz

if TYPE_CHECKING:
    from eodag.api.product import EOProduct

from eodag.plugins.crunch.base import Crunch

logger = logging.getLogger("eodag.crunch.date")


class FilterDate(Crunch):
    """FilterDate cruncher: filter products by date

    Allows to filter out products that are older than a start date (optional) or more recent than an end date
    (optional).

    :param config: Crunch configuration, may contain :

        * ``start`` (``str``): start sensing time in iso format
        * ``end`` (``str``): end sensing time in iso format
    """

    @staticmethod
    def sort_product_by_start_date(product: EOProduct) -> dt:
        """Get product start date"""
        start_date = product.properties.get("startTimeFromAscendingNode")
        if not start_date:
            # Retrieve year, month, day, hour, minute, second of EPOCH start
            epoch = time.gmtime(0)[:-3]
            start_date = datetime.datetime(*epoch).isoformat()
        return dateutil.parser.parse(start_date)

    def proceed(
        self, products: list[EOProduct], **search_params: Any
    ) -> list[EOProduct]:
        """Execute crunch: Filter products between start and end dates.

        :param products: A list of products resulting from a search
        :returns: The filtered products
        """
        logger.debug("Start filtering by date")
        if not products:
            return []

        # filter start date
        filter_start_str = self.config.__dict__.get("start")
        if filter_start_str:
            filter_start = dateutil.parser.parse(filter_start_str)
            if not filter_start.tzinfo:
                filter_start = filter_start.replace(tzinfo=tz.UTC)
        else:
            filter_start = None

        # filter end date
        filter_end_str = self.config.__dict__.get("end")
        if filter_end_str:
            filter_end = dateutil.parser.parse(filter_end_str)
            if not filter_end.tzinfo:
                filter_end = filter_end.replace(tzinfo=tz.UTC)
        else:
            filter_end = None

        if not filter_start and not filter_end:
            return products

        filtered: list[EOProduct] = []
        for product in products:

            # product start date
            product_start_str = product.properties.get("startTimeFromAscendingNode")
            if product_start_str:
                product_start = dateutil.parser.parse(product_start_str)
                if not product_start.tzinfo:
                    product_start = product_start.replace(tzinfo=tz.UTC)
            else:
                product_start = None

            # product end date
            product_end_str = product.properties.get("completionTimeFromAscendingNode")
            if product_end_str:
                product_end = dateutil.parser.parse(product_end_str)
                if not product_end.tzinfo:
                    product_end = product_end.replace(tzinfo=tz.UTC)
            else:
                product_end = None

            if filter_start and product_start and product_start < filter_start:
                continue
            if filter_end and product_end and product_end > filter_end:
                continue
            if filter_end and product_start and product_start > filter_end:
                continue

            filtered.append(product)
        logger.info("Finished filtering products. %s resulting products", len(filtered))
        return filtered



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Optional, cast

import boto3
from botocore.exceptions import ClientError, ProfileNotFound
from botocore.handlers import disable_signing

from eodag.api.product._assets import Asset
from eodag.plugins.authentication.base import Authentication
from eodag.types import S3SessionKwargs
from eodag.utils import get_bucket_name_and_prefix
from eodag.utils.exceptions import AuthenticationError, EodagError

if TYPE_CHECKING:
    from mypy_boto3_s3 import S3Client, S3ServiceResource
    from mypy_boto3_s3.service_resource import BucketObjectsCollection

    from eodag.config import PluginConfig


logger = logging.getLogger("eodag.download.aws_auth")

AWS_AUTH_ERROR_MESSAGES = [
    "AccessDenied",
    "InvalidAccessKeyId",
    "SignatureDoesNotMatch",
    "InvalidRequest",
]


def raise_if_auth_error(exception: ClientError, provider: str) -> None:
    """Raises an error if given exception is an authentication error"""
    err = cast(dict[str, str], exception.response["Error"])
    if err["Code"] in AWS_AUTH_ERROR_MESSAGES and "key" in err["Message"].lower():
        raise AuthenticationError(
            f"Please check your credentials for {provider}.",
            f"HTTP Error {exception.response['ResponseMetadata']['HTTPStatusCode']} returned.",
            err["Code"] + ": " + err["Message"],
        )


def create_s3_session(**kwargs) -> boto3.Session:
    """create s3 session based on available credentials

    :param kwargs: keyword arguments containing credentials
    :returns: boto3 Session
    """
    try:
        s3_session = boto3.Session(**kwargs)
    except ProfileNotFound:
        raise AuthenticationError(
            f"AWS profile {kwargs['profile_name']} not found, please check your credentials configuration"
        )
    return s3_session


class AwsAuth(Authentication):
    """AWS authentication plugin

    The authentication method will be chosen depending on which parameters are available in the configuration:

    * auth using ``profile_name`` (if credentials are given and contain ``aws_profile``)
    * auth using ``aws_access_key_id``, ``aws_secret_access_key`` and optionally ``aws_session_token``
      (if credentials are given but no ``aws_profile``)
    * auth using current environment - AWS environment variables and/or ``~/.aws/*``
      (if no credentials are given in config)
    * auth anonymously using no-sign-request if no credentials are given in config and
      auth using current environment failed

    :param provider: provider name
    :param config: Authentication plugin configuration:

        * :attr:`~eodag.config.PluginConfig.type` (``str``) (**mandatory**): AwsAuth
        * :attr:`~eodag.config.PluginConfig.auth_error_code` (``int``) (mandatory for ``creodias_s3``):
          which error code is returned in case of an authentication error
        * :attr:`~eodag.config.PluginConfig.s3_endpoint` (``str``): s3 endpoint url
        * :attr:`~eodag.config.PluginConfig.requester_pays` (``bool``): whether download is done
          from a requester-pays bucket or not; default: ``False``

    """

    def __init__(self, provider: str, config: PluginConfig) -> None:
        super(AwsAuth, self).__init__(provider, config)
        self.s3_session: Optional[boto3.Session] = None
        self.s3_resource: Optional[S3ServiceResource] = None
        # set default for requester_pays if not given
        self.config.__dict__.setdefault("requester_pays", False)

    def _create_s3_session_from_credentials(self) -> boto3.Session:
        credentials = getattr(self.config, "credentials", {}) or {}
        if "aws_profile" in credentials:
            logger.debug("Authentication using AWS profile")
            return create_s3_session(profile_name=credentials["aws_profile"])
        # auth using aws keys
        elif credentials.get("aws_access_key_id") and credentials.get(
            "aws_secret_access_key"
        ):
            s3_session_kwargs: S3SessionKwargs = {
                "aws_access_key_id": credentials["aws_access_key_id"],
                "aws_secret_access_key": credentials["aws_secret_access_key"],
            }
            if credentials.get("aws_session_token"):
                s3_session_kwargs["aws_session_token"] = credentials[
                    "aws_session_token"
                ]
            return create_s3_session(**s3_session_kwargs)
        else:
            # auth using env variables or ~/.aws
            logger.debug("Authentication using AWS environment")
            return create_s3_session()

    def _create_s3_resource(self) -> S3ServiceResource:
        """create s3 resource based on s3 session"""
        if not self.s3_session:
            self.s3_session = self._create_s3_session_from_credentials()
        endpoint_url = getattr(self.config, "s3_endpoint", None)
        if self.s3_session.get_credentials():
            return self.s3_session.resource(
                service_name="s3",
                endpoint_url=endpoint_url,
            )
        # could not auth using credentials: use no-sign-request strategy
        logger.debug(
            "Authentication using AWS no-sign-request strategy (no credentials found)"
        )
        s3_resource = boto3.resource(service_name="s3", endpoint_url=endpoint_url)
        s3_resource.meta.client.meta.events.register(
            "choose-signer.s3.*", disable_signing
        )
        return s3_resource

    def get_s3_client(self) -> S3Client:
        """Get S3 client from S3 resource

        :returns: boto3 client
        """
        if not self.s3_resource:
            self.s3_resource = self._create_s3_resource()
        return self.s3_resource.meta.client

    def authenticate(self) -> S3ServiceResource:
        """Authenticate

        :returns: S3 Resource created based on an S3 session
        """
        self.s3_resource = self._create_s3_resource()
        return self.s3_resource

    def _get_authenticated_objects(
        self, bucket_name: str, prefix: str
    ) -> BucketObjectsCollection:
        """Get boto3 authenticated objects for the given bucket

        :param bucket_name: Bucket containg objects
        :param prefix: Prefix used to filter objects
        :returns: The boto3 authenticated objects
        """
        if not self.s3_resource:
            self.s3_resource = self._create_s3_resource()
        try:
            if self.config.requester_pays:
                objects = self.s3_resource.Bucket(bucket_name).objects.filter(
                    RequestPayer="requester"
                )
            else:
                objects = self.s3_resource.Bucket(bucket_name).objects
            list(objects.filter(Prefix=prefix).limit(1))
            if objects:
                logger.debug(
                    "Authentication for bucket %s succeeded; returning available objects",
                    bucket_name,
                )
                return objects
        except ClientError as e:
            if e.response.get("Error", {}).get("Code", {}) in AWS_AUTH_ERROR_MESSAGES:
                pass
            else:
                raise e
        logger.debug(
            "Authentication for bucket %s failed, please check the credentials",
            bucket_name,
        )

        raise AuthenticationError(
            "Unable do authenticate on s3://%s using credendials configuration"
            % bucket_name
        )

    def authenticate_objects(
        self,
        bucket_names_and_prefixes: list[tuple[str, Optional[str]]],
    ) -> dict[str, BucketObjectsCollection]:
        """
        Authenticates with s3 and retrieves the available objects

        :param bucket_names_and_prefixes: list of bucket names and corresponding path prefixes
        :raises AuthenticationError: authentication is not possible
        :return: authenticated objects per bucket
        """

        authenticated_objects: dict[str, Any] = {}
        auth_error_messages: set[str] = set()
        for _, pack in enumerate(bucket_names_and_prefixes):

            bucket_name, prefix = pack
            if not prefix:
                continue
            if bucket_name not in authenticated_objects:
                # get Prefixes longest common base path
                common_prefix = ""
                prefix_split = prefix.split("/")
                prefixes_in_bucket = len(
                    [p for b, p in bucket_names_and_prefixes if b == bucket_name]
                )
                for i in range(1, len(prefix_split)):
                    common_prefix = "/".join(prefix_split[0:i])
                    if (
                        len(
                            [
                                p
                                for b, p in bucket_names_and_prefixes
                                if p and b == bucket_name and common_prefix in p
                            ]
                        )
                        < prefixes_in_bucket
                    ):
                        common_prefix = "/".join(prefix_split[0 : i - 1])
                        break
                try:
                    # connect to aws s3 and get bucket auhenticated objects
                    authenticated_objects[
                        bucket_name
                    ] = self._get_authenticated_objects(bucket_name, common_prefix)

                except AuthenticationError as e:
                    logger.warning("Unexpected error: %s" % e)
                    logger.warning("Skipping %s/%s" % (bucket_name, prefix))
                    auth_error_messages.add(str(e))
                except ClientError as e:
                    raise_if_auth_error(e, self.provider)
                    logger.warning("Unexpected error: %s" % e)
                    logger.warning("Skipping %s/%s" % (bucket_name, prefix))
                    auth_error_messages.add(str(e))

        # could not auth on any bucket
        if not authenticated_objects:
            raise AuthenticationError(", ".join(auth_error_messages))
        return authenticated_objects

    def get_rio_env(self) -> dict[str, Any]:
        """Get rasterio environment variables needed for data access authentication.

        :returns: The rasterio environement variables
        """
        rio_env_kwargs = {}
        if endpoint_url := getattr(self.config, "s3_endpoint", None):
            rio_env_kwargs["endpoint_url"] = endpoint_url.split("://")[-1]

        if self.s3_session is None:
            self.authenticate()

        if self.config.requester_pays:
            rio_env_kwargs["requester_pays"] = True

        return {
            "session": self.s3_session,
            **rio_env_kwargs,
        }

    def presign_url(
        self,
        asset: Asset,
        expires_in: int = 3600,
    ) -> str:
        """This method is used to presign a url to download an asset from S3.

        :param asset: asset for which the url shall be presigned
        :param expires_in: expiration time of the presigned url in seconds
        :returns: presigned url
        :raises: :class:`~eodag.utils.exceptions.EodagError`
        :raises: :class:`NotImplementedError`
        """
        if not getattr(self.config, "support_presign_url", True):
            raise NotImplementedError(
                f"presign_url is not supported for provider {self.provider}"
            )

        s3_client = self.get_s3_client()
        bucket, prefix = get_bucket_name_and_prefix(asset["href"])
        try:
            presigned_url = s3_client.generate_presigned_url(
                "get_object",
                Params={
                    "Bucket": bucket,
                    "Key": prefix,
                },
                ExpiresIn=expires_in,
            )
            return presigned_url
        except ClientError:
            raise EodagError(f"Couldn't get a presigned URL for '{asset}'.")



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""EODAG product package"""

from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:
    from eodag.plugins.manager import PluginManager

try:
    # import from eodag-cube if installed
    from eodag_cube.api.product import (  # pyright: ignore[reportMissingImports]
        Asset,
        AssetsDict,
        EOProduct,
    )
except ImportError:
    from ._assets import Asset, AssetsDict  # type: ignore[assignment]
    from ._product import EOProduct  # type: ignore[assignment]

# exportable content
__all__ = ["Asset", "AssetsDict", "EOProduct"]


def unregistered_product_from_item(
    feature: dict[str, Any], provider: str, plugins_manager: "PluginManager"
) -> Optional[EOProduct]:
    """Create an EOProduct from a STAC item, map its metadata, but without registering its plugins.

    :param feature: The STAC item to convert into an EOProduct.
    :param provider: The associated provider from which configuration should be used for mapping.
    :param plugins_manager: The plugins manager instance to use for retrieving search plugins.
    :returns: An EOProduct instance if the item can be normalized, otherwise None.
    """
    for search_plugin in plugins_manager.get_search_plugins(provider=provider):
        if hasattr(search_plugin, "normalize_results"):
            products = search_plugin.normalize_results([feature])
            if len(products) > 0:
                # properties cleanup
                for prop in ("start_datetime", "end_datetime"):
                    products[0].properties.pop(prop, None)
                # set product type if not already set
                if products[0].product_type is None:
                    products[0].product_type = products[0].properties.get("productType")
                return products[0]
    return None
